#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

#define BLOCK_SIZE 16

// CUDA kernel implementing GEMM with shared memory tiling and explicit loop unrolling
__global__ void gemm_unrolled_kernel(
    const float* __restrict__ x,       // Input: [batch_size x in_features]
    const float* __restrict__ weight,  // Weight: [out_features x in_features]
    const float* __restrict__ bias,    // Bias: [out_features]
    float* __restrict__ output,        // Output: [batch_size x out_features]
    const int batch_size,
    const int in_features,
    const int out_features,
    const float multiplier,
    const float negative_slope
) {
    // Calculate global row and column indices
    int row = blockIdx.x * BLOCK_SIZE + threadIdx.x;
    int col = blockIdx.y * BLOCK_SIZE + threadIdx.y;
    float sum = 0.0f;

    // Shared memory tiles for x and weight, with added padding to avoid bank conflicts
    __shared__ float s_x[BLOCK_SIZE][BLOCK_SIZE + 1];
    __shared__ float s_w[BLOCK_SIZE][BLOCK_SIZE + 1];

    // Determine the number of tiles needed along the in_features dimension
    int numTiles = (in_features + BLOCK_SIZE - 1) / BLOCK_SIZE;

    // Unroll the tiling loop to reduce loop overhead
    #pragma unroll
    for (int t = 0; t < numTiles; t++) {
        int tiled_index = t * BLOCK_SIZE;

        // Load a tile from input x into shared memory, use __ldg for potential cache load hints
        if (row < batch_size && (tiled_index + threadIdx.y) < in_features) {
            s_x[threadIdx.x][threadIdx.y] = __ldg(&x[row * in_features + tiled_index + threadIdx.y]);
        } else {
            s_x[threadIdx.x][threadIdx.y] = 0.0f;
        }

        // Load a tile from weight, transpose it for better coalescing
        if (col < out_features && (tiled_index + threadIdx.x) < in_features) {
            s_w[threadIdx.y][threadIdx.x] = __ldg(&weight[col * in_features + tiled_index + threadIdx.x]);
        } else {
            s_w[threadIdx.y][threadIdx.x] = 0.0f;
        }

        __syncthreads();

        // Unroll the inner loop fully given the BLOCK_SIZE is known at compile time
        #pragma unroll
        for (int k = 0; k < BLOCK_SIZE; k++) {
            sum = __fmaf_rn(s_x[threadIdx.x][k], s_w[threadIdx.y][k], sum);
        }

        __syncthreads();
    }

    // Write the computed value to the output, fusing bias addition, scaling and LeakyReLU
    if (row < batch_size && col < out_features) {
        float result = (sum + bias[col]) * multiplier;
        result = (result > 0.0f) ? result : result * negative_slope;
        output[row * out_features + col] = result;
    }
}

// Host function to set up kernel execution
torch::Tensor module_fn_forward(
    torch::Tensor x,
    float multiplier,
    float negative_slope,
    torch::Tensor weight,
    torch::Tensor bias
) {
    TORCH_CHECK(x.device().is_cuda(), "x must be a CUDA tensor");
    TORCH_CHECK(weight.device().is_cuda(), "weight must be a CUDA tensor");
    TORCH_CHECK(bias.device().is_cuda(), "bias must be a CUDA tensor");

    const int batch_size = x.size(0);
    const int in_features = x.size(1);
    const int out_features = weight.size(0);

    TORCH_CHECK(weight.size(1) == in_features, "Weight in_features must match x in_features");
    TORCH_CHECK(bias.size(0) == out_features, "Bias size must match weight out_features");

    auto output = torch::zeros({batch_size, out_features}, x.options());

    dim3 block(BLOCK_SIZE, BLOCK_SIZE);
    dim3 grid(
        (batch_size + BLOCK_SIZE - 1) / BLOCK_SIZE,
        (out_features + BLOCK_SIZE - 1) / BLOCK_SIZE
    );

    gemm_unrolled_kernel<<<grid, block>>>(
        x.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.data_ptr<float>(),
        output.data_ptr<float>(),
        batch_size,
        in_features,
        out_features,
        multiplier,
        negative_slope
    );

    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &module_fn_forward, "GEMM forward CUDA kernel with explicit loop unrolling");
}
